﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data.SQLite;
using System.Text;
using ThinkGeo.MapSuite.Core;
using System.Linq;
using WorldMapKitDataExtractor.Model;

namespace WorldMapKitDataExtractor
{
    public class Extractor
    {
        private string inputDataPath;
        private string outputDataPath;
        private string inputDataPrj;
        private string boundaryPrj;
        private bool preserveCountryLevelData;

        public Extractor(string inputDataPath, string outputDataPath, string inputDataPrj = null, string boundaryPrj = null, bool preserveCountryLevelData = true)
        {
            this.inputDataPath = inputDataPath;
            this.outputDataPath = outputDataPath;
            this.inputDataPrj = inputDataPrj;
            this.boundaryPrj = boundaryPrj;
            this.preserveCountryLevelData = preserveCountryLevelData;
        }

        public Action<string> UpdateStatus { get; set; }

        public string InputDataPath
        {
            get { return inputDataPath; }
            set { inputDataPath = value; }
        }

        public string OutputDataPath
        {
            get { return outputDataPath; }
            set { outputDataPath = value; }
        }

        public string InputDataPrj
        {
            get { return inputDataPrj; }
            set { inputDataPrj = value; }
        }

        public string BoundaryPrj
        {
            get { return boundaryPrj; }
            set { boundaryPrj = value; }
        }

        public bool PreserveCountryLevelData
        {
            get { return preserveCountryLevelData; }
            set { preserveCountryLevelData = value; }
        }

        public void ExtractDataByShape(Collection<BoundingBoxWithZoomLevels> boundaryFeatureSet)
        {
            UpdateStatus("Beginning to extract world map kit data by shape file.");

            Collection<BoundingBoxWithZoomLevels> projectedBoundaryFeaturesWithZoomLevels = new Collection<BoundingBoxWithZoomLevels>();
            Collection<RectangleShape> projectedBoundingBoxes = new Collection<RectangleShape>();
            foreach (BoundingBoxWithZoomLevels boundaryFeatureWithZoomLevels in boundaryFeatureSet)
            {
                RectangleShape projectedBoundaryFeature = ProjectBoundingBox(boundaryFeatureWithZoomLevels.BoundingBox);
                projectedBoundaryFeaturesWithZoomLevels.Add(new BoundingBoxWithZoomLevels(projectedBoundaryFeature, boundaryFeatureWithZoomLevels.StartingZoomLevel, boundaryFeatureWithZoomLevels.EndingZoomLevel));
                projectedBoundingBoxes.Add(projectedBoundaryFeature);
            }

            SQLiteConnection.CreateFile(outputDataPath);
            SQLiteConnection targetDatabaseConnection = new SQLiteConnection($"Data Source={outputDataPath};Version=3;");
            targetDatabaseConnection.Open();
            SQLiteConnection sourceDatabaseConnection = new SQLiteConnection($"Data Source={inputDataPath};Version=3;");
            sourceDatabaseConnection.Open();

            //this is the set of layers we use to determine which tables to grab for each bounding area
            OsmWorldMapKitLayer wmkLayer = new OsmWorldMapKitLayer(sourceDatabaseConnection.ConnectionString, OsmWorldMapKitDatabaseType.Sqlite);
            wmkLayer.Open();

            CopyDataToABuffer(GetBoundingBoxesForLayers(boundaryFeatureSet, sourceDatabaseConnection, wmkLayer.Layers), sourceDatabaseConnection);
            ProcessBaselandByShapes(projectedBoundingBoxes, targetDatabaseConnection);

            UpdateStatus("Executing vacuum for output database.");
            ExecuteNonQueryCommand("VACUUM;", targetDatabaseConnection);

            targetDatabaseConnection.Close();
            targetDatabaseConnection.Dispose();
            UpdateStatus("Done extracting world map kit data!");
        }

        private void CopyDataToABuffer(Dictionary<FeatureLayer, Collection<RectangleShape>> layersWithBoundingShapes, SQLiteConnection sourceDatabaseConnection)
        {
            Proj4Projection proj4 = new Proj4Projection(inputDataPrj, inputDataPrj);
            GeographyUnit geographyUnit = proj4.GetInternalGeographyUnit();

            //OSM WorldMapKit contains several views and view indices that we want to grab as well for our OSMLayer, so we create temporary tables of the data from these within the boundingbox

            var views = GetViews(sourceDatabaseConnection);

            foreach (View view in views)
            {
                if (layersWithBoundingShapes.Keys.Where(featureLayer => featureLayer.Name == view.Name).Count() > 0)
                {
                    Collection<RectangleShape> boundingBoxes = layersWithBoundingShapes[layersWithBoundingShapes.Keys.Where(featureLayer => featureLayer.Name == view.Name).ElementAt(0)];
                    CreateTempTableForView(view, boundingBoxes, sourceDatabaseConnection);
                }
                else
                {
                    //create the table so the database is valid
                    CreateTempTableForView(view, new Collection<RectangleShape>() { new RectangleShape(0, 0, 0, 0) }, sourceDatabaseConnection);
                }
            }

            //get the features from every table within the bounding box and add them to the temporary output file

            var tables = GetTables(sourceDatabaseConnection);
            foreach (var table in tables)
            {
                UpdateStatus($"Copying {table.TableName} table to output database.");
                SqliteFeatureSource source = new SqliteFeatureSource($@"Data Source={inputDataPath};Version=3;", table.TableName);
                source.Open();

                Collection<Feature> intersectFeatures = new Collection<Feature>();

                if (layersWithBoundingShapes.Keys.Where(featureLayer => featureLayer.Name == table.TableName).Count() > 0)
                {
                    intersectFeatures = GetFeaturesIntersects(source, layersWithBoundingShapes[layersWithBoundingShapes.Keys.Where(featureLayer => featureLayer.Name == table.TableName).ElementAt(0)], ReturningColumnsType.AllColumns);
                }

                Collection<string> targetTables = SqliteFeatureSource.GetTableNames($@"Data Source={outputDataPath};Version=3;");

                if (!targetTables.Contains(table.TableName))
                    SqliteFeatureSource.CreateTable($@"Data Source={outputDataPath};Version=3;", table.TableName, table.Columns, geographyUnit);

                SqliteFeatureSource tempSource = new SqliteFeatureSource($@"Data Source={outputDataPath};Version=3;", table.TableName);
                tempSource.Open();
                tempSource.BeginTransaction();

                int counter = 0;
                UpdateStatus($"Adding <1000 new features.");
                foreach (var feature in intersectFeatures.Distinct(new FeatureComparer()))
                {
                    tempSource.AddFeature(feature);
                    counter++;
                    if (counter % 1000 == 0)
                    {
                        UpdateStatus($"Adding {counter} new features.");
                    }
                }
                UpdateStatus("Commiting sqlite database transaction.");
                tempSource.CommitTransaction();
                tempSource.Close();
            }
        }

        private void ProcessBaselandByShapes(Collection<RectangleShape> boundingBoxes, SQLiteConnection connection)
        {
            UpdateStatus("Processing baseland by shapes.");
            if (!preserveCountryLevelData)
                return;

            //remove full world bounding box from boundary features
            boundingBoxes.RemoveAt(0);

            SqliteFeatureSource baselandFeatureSource = new SqliteFeatureSource(connection.ConnectionString, "osm_baseland_polygon", "id", "geometry");
            SqliteFeatureSource baseland1mFeatureSource = new SqliteFeatureSource(connection.ConnectionString, "osm_baseland1m_polygon", "id", "geometry");

            baselandFeatureSource.Open();
            baseland1mFeatureSource.Open();

            UpdateStatus("Querying baseland features from osm_baseland_polygon table.");
            List<Feature> baselandFeatures = new List<Feature>();
            foreach (var boundingBox in boundingBoxes)
            {
                Collection<Feature> features = baselandFeatureSource.SpatialQuery(boundingBox, QueryType.Intersects, ReturningColumnsType.NoColumns);
                baselandFeatures.AddRange(features);
            }

            UpdateStatus("Querying baseland1m features from osm_baseland1m_polygon table.");
            Collection<Feature> baseland1mFeatures = baseland1mFeatureSource.GetAllFeatures(ReturningColumnsType.NoColumns);

            if (baselandFeatures.Count == 0 || baseland1mFeatures.Count == 0)
            {
                UpdateStatus("Cannot process baseland by shapes - one or more tables is empty");
                return;
            }

            Collection<Feature> newBaselandFeatures = new Collection<Feature>();

            int counter = 0;
            UpdateStatus($"Getting <1000 intersection features for baseland feature.");
            foreach (Feature feature in baselandFeatures)
            {
                Feature validFeature = SqlTypesGeometryHelper.MakeValid(feature);
                Feature intersectionFeature = GetInterSection(validFeature, boundingBoxes);
                if (!(intersectionFeature == null))
                {
                    newBaselandFeatures.Add(intersectionFeature);
                    counter++;
                    if (counter % 1000 == 0)
                    {
                        UpdateStatus($"Getting {counter} intersection features for baseland feature.");
                    }
                }
            }

            counter = 0;
            UpdateStatus($"Getting <1000 difference features for baseland1m feature.");
            foreach (Feature feature in baseland1mFeatures)
            {
                if (IsDisjointed(feature, boundingBoxes))
                {
                    newBaselandFeatures.Add(feature);
                    counter++;
                }
                else
                {
                    Feature validFeature = SqlTypesGeometryHelper.MakeValid(feature);
                    Feature difference = GetDifference(validFeature, boundingBoxes);

                    if (difference != null)
                    {
                        newBaselandFeatures.Add(difference);
                        counter++;
                    }
                }
                if (counter != 0 && counter % 1000 == 0)
                {
                    UpdateStatus($"Getting {counter} difference features for baseland1m feature.");
                }
            }

            baselandFeatureSource.BeginTransaction();
            UpdateStatus("Deleting table osm_baseland_polygon.");
            baselandFeatureSource.ExecuteNonQuery("DELETE FROM osm_baseland_polygon;");
            UpdateStatus("Deleting table idx_osm_baseland_polygon_geometry.");
            baselandFeatureSource.ExecuteNonQuery("DELETE FROM idx_osm_baseland_polygon_geometry;");

            counter = 0;
            UpdateStatus($"Adding <1000 new features.");
            foreach (Feature f in newBaselandFeatures)
            {
                Feature feature = new Feature(f.GetWellKnownBinary());
                baselandFeatureSource.AddFeature(feature);
                counter++;
                if (counter % 1000 == 0)
                {
                    UpdateStatus($"Adding {counter} new features.");
                }
            }
            UpdateStatus("Commiting sqlite database transaction.");
            baselandFeatureSource.CommitTransaction();
        }

        private Collection<View> GetViews(SQLiteConnection connection)
        {
            // Here we query all of the views
            SQLiteCommand getTablesCommand = new SQLiteCommand(@"SELECT name, sql FROM sqlite_master WHERE (type='view');", connection);
            SQLiteDataReader getTablesreader = getTablesCommand.ExecuteReader();

            Collection<View> views = new Collection<View>();

            // Add the views into a collection
            while (getTablesreader.Read())
            {
                string viewName = getTablesreader.GetValue(0).ToString();
                string createStatement = getTablesreader.GetValue(1).ToString();

                //Extract the base table from the create statement
                //Get rid of new lines
                string baseTable = createStatement.Replace("\n", "");
                //Get rid of tabs
                baseTable = baseTable.Replace("\t", "");
                //Get only the text after the "FROM" Statement
                baseTable = baseTable.Substring(baseTable.LastIndexOf("FROM ") + "FROM ".Length);
                //Get only the text before tho "WHERE" statement
                //uses inline if to handle the case of no "WHERE" statement
                //see osm_address_point for an example of this
                baseTable = baseTable.Substring(0, baseTable.LastIndexOf("WHERE") > 0 ? baseTable.LastIndexOf("WHERE") : baseTable.Length);
                //Get rid of any leading or trailing spaces that might exist
                baseTable = baseTable.Trim();
                View view = new View(viewName, baseTable, createStatement);

                views.Add(view);
            }
            return views;
        }

        private Collection<Table> GetTables(SQLiteConnection connection)
        {
            // Here we query all of the tables which are not master tables or the indexes
            SQLiteCommand getTablesCommand = new SQLiteCommand(@"SELECT sqlite_master.name FROM sqlite_master LEFT OUTER JOIN sqlite_temp_master ON sqlite_master.name = sqlite_temp_master.name WHERE (sqlite_master.type = 'table' OR sqlite_temp_master.type = 'table') AND sqlite_master.name NOT LIKE 'sqlite_%' and sqlite_master.name NOT LIKE 'idx_%' and sqlite_master.name NOT LIKE 'planet_%'", connection);
            SQLiteDataReader getTablesreader = getTablesCommand.ExecuteReader();

            Collection<Table> tables = new Collection<Table>();

            // Add the tables into a collection of table objects
            while (getTablesreader.Read())
            {
                Table table = new Table();
                table.TableName = getTablesreader.GetValue(0).ToString();
                table.Columns = new Collection<SqliteColumn>();
                tables.Add(table);
            }

            // Cleanup the table reader and command
            getTablesreader.Close();
            getTablesreader.Dispose();

            // For each table we now get and load the column names and data types
            foreach (Table table in tables)
            {
                // Get the column info for the table
                SQLiteCommand getTableSchemasCommand = new SQLiteCommand(string.Format("Pragma table_info({0});", table.TableName), connection);
                SQLiteDataReader getSchemaReader = getTableSchemasCommand.ExecuteReader();

                while (getSchemaReader.Read())
                {
                    string columnName = getSchemaReader.GetValue(1).ToString();
                    string columnType = getSchemaReader.GetValue(2).ToString();

                    if (!(columnName.Equals("geometry") || columnName.Equals("id")))
                    {
                        switch (columnType)
                        {
                            case "character varying":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Text });
                                break;
                            case "bigint":
                            case "INT":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Integer });
                                break;
                            case "double precision":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Real });
                                break;
                            case "numeric":
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = SqliteColumnType.Text });
                                break;
                            default:
                                table.Columns.Add(new SqliteColumn() { ColumnName = columnName, ColumnType = (SqliteColumnType)Enum.Parse(typeof(SqliteColumnType), columnType, true) });
                                break;
                        }
                    }
                }

                // Cleanup the column command and reader
                getSchemaReader.Close();
                getSchemaReader.Dispose();
            }

            return tables;
        }

        private void CreateTempTableForView(View view, Collection<RectangleShape> boundingBoxes, SQLiteConnection sourceDatabaseConnection)
        {
            UpdateStatus($"Creating temporary table for {view.Name}");

            StringBuilder createNewTempTableFromViewCommand = new StringBuilder();

            createNewTempTableFromViewCommand.AppendFormat("CREATE TEMP TABLE {0} AS ", view.Name);
            createNewTempTableFromViewCommand.AppendFormat("SELECT * FROM {0} WHERE id in (", view.Name);
            createNewTempTableFromViewCommand.AppendFormat("SELECT id FROM {0} WHERE id in (", view.BaseTable);
            createNewTempTableFromViewCommand.AppendFormat("SELECT id from idx_{0}_geometry WHERE (maxx < {1} AND maxy < {2} AND minx > {3} AND miny > {4})", view.Name, boundingBoxes[0].UpperRightPoint.X, boundingBoxes[0].UpperRightPoint.Y, boundingBoxes[0].LowerLeftPoint.X, boundingBoxes[0].LowerLeftPoint.Y);
            for (int i = 1; i < boundingBoxes.Count; i++)
            {
                createNewTempTableFromViewCommand.AppendFormat(" OR (maxx < {0} AND maxy < {1} AND minx > {2} AND miny > {3})", boundingBoxes[i].UpperRightPoint.X, boundingBoxes[i].UpperRightPoint.Y, boundingBoxes[i].LowerLeftPoint.X, boundingBoxes[i].LowerLeftPoint.Y);
            }

            createNewTempTableFromViewCommand.AppendFormat("));");

            ExecuteNonQueryCommand(createNewTempTableFromViewCommand.ToString(), sourceDatabaseConnection);
        }

        private void ExecuteNonQueryCommand(StringBuilder commandText, SQLiteConnection connection)
        {
            ExecuteNonQueryCommand(commandText.ToString(), connection);
        }

        private void ExecuteNonQueryCommand(string commandText, SQLiteConnection connection)
        {
            SQLiteCommand command = new SQLiteCommand(commandText, connection);
            command.ExecuteNonQuery();
            command.Dispose();
        }

        private RectangleShape ProjectBoundingBox(RectangleShape boundingBox)
        {
            Proj4Projection proj = new Proj4Projection(boundaryPrj, inputDataPrj);
            proj.Open();
            RectangleShape projectedBoundaryFeature = proj.ConvertToExternalProjection(boundingBox);
            proj.Close();
            return projectedBoundaryFeature;
        }

        private Collection<Feature> GetFeaturesIntersects(FeatureSource source, Collection<RectangleShape> boundingBoxes, ReturningColumnsType returningColumnsType)
        {
            List<Feature> results = new List<Feature>();
            foreach (var boundingBox in boundingBoxes)
            {
                Collection<Feature> allPossibleFeatures = source.GetFeaturesInsideBoundingBox(boundingBox, returningColumnsType);

                Collection<Feature> returnFeatures = new Collection<Feature>();
                foreach (Feature feature in allPossibleFeatures)
                {
                    ShapeValidationResult validationResult = feature.GetShape().Validate(ShapeValidationMode.Simple);

                    if (validationResult.IsValid && feature.GetShape().Intersects(boundingBox))
                    {
                        returnFeatures.Add(feature);
                    }
                }
                results.AddRange(returnFeatures);
            }

            return new Collection<Feature>(results);
        }

        private Dictionary<FeatureLayer, Collection<RectangleShape>> GetBoundingBoxesForLayers(Collection<BoundingBoxWithZoomLevels> boundaryFeatures, SQLiteConnection sourceDatabaseConnection, Collection<FeatureLayer> featureLayers)
        {
            Dictionary<FeatureLayer, Collection<RectangleShape>> layersWithBoundingBoxes = new Dictionary<FeatureLayer, Collection<RectangleShape>>();

            foreach (BoundingBoxWithZoomLevels boundingBoxParameters in boundaryFeatures)
            {
                Collection<FeatureLayer> layersForBoundingBoxes = GetAllFeatureLayers(featureLayers, boundingBoxParameters.StartingZoomLevel, boundingBoxParameters.EndingZoomLevel);

                foreach (FeatureLayer featureLayer in layersForBoundingBoxes)
                {
                    if (!layersWithBoundingBoxes.Keys.Contains(featureLayer))
                    {
                        var boxes = new Collection<RectangleShape>();
                        boxes.Add(boundingBoxParameters.BoundingBox);
                        layersWithBoundingBoxes.Add(featureLayer, boxes);
                    }
                    else
                    {
                        layersWithBoundingBoxes[featureLayer].Add(boundingBoxParameters.BoundingBox);
                    }
                }
            }

            return layersWithBoundingBoxes;
        }

        private Collection<FeatureLayer> GetAllFeatureLayers(Collection<FeatureLayer> featureLayers, int fromZoomLevelNumber, int toZoomLevelNumber)
        {
            Collection<FeatureLayer> featureLayersForZoomLevels = new Collection<FeatureLayer>();
            Dictionary<int, Collection<FeatureLayer>> allZoomLevelFeatureLayers = GetAllFeatureLayersCore(featureLayers);

            for (int i = fromZoomLevelNumber; i <= toZoomLevelNumber; i++)
            {
                foreach (FeatureLayer featureLayer in allZoomLevelFeatureLayers[i])
                {
                    if (!featureLayersForZoomLevels.Contains(featureLayer))
                    {
                        featureLayersForZoomLevels.Add(featureLayer);
                    }
                }
            }
            return featureLayersForZoomLevels;
        }

        private Dictionary<int, Collection<FeatureLayer>> GetAllFeatureLayersCore(Collection<FeatureLayer> featureLayers)
        {
            Dictionary<int, Collection<FeatureLayer>> featureLayersInAllZoomLevels = new Dictionary<int, Collection<FeatureLayer>>();
            for (int i = 1; i <= 20; i++)
            {
                featureLayersInAllZoomLevels.Add(i, new Collection<FeatureLayer>());
            }

            foreach (FeatureLayer layer in featureLayers)
            {
                Collection<ZoomLevel> layerZoomLevels = layer.ZoomLevelSet.GetZoomLevels();

                for (int i = 1; i <= 20; i++)
                {
                    ZoomLevel zoomLevel = layerZoomLevels[i - 1];

                    if (zoomLevel.CustomStyles.Count != 0 || !IsDefaultStyleSet(zoomLevel))
                    {
                        if ((int)zoomLevel.ApplyUntilZoomLevel == 0)
                        {
                            if (!featureLayersInAllZoomLevels[i].Contains(layer))
                            {
                                featureLayersInAllZoomLevels[i].Add(layer);
                            }
                        }
                        else
                        {
                            for (int j = i; j <= (int)zoomLevel.ApplyUntilZoomLevel; j++)
                            {
                                if (!featureLayersInAllZoomLevels[j].Contains(layer))
                                {
                                    featureLayersInAllZoomLevels[j].Add(layer);
                                }
                            }
                        }
                    }
                }
            }

            return featureLayersInAllZoomLevels;
        }

        private bool IsDefaultStyleSet(ZoomLevel zoomLevel)
        {
            return IsDefaultAreaStyle(zoomLevel.DefaultAreaStyle) && IsDefaultLineStyle(zoomLevel.DefaultLineStyle) && IsDefaultPointStyle(zoomLevel.DefaultPointStyle) && IsDefaultTextStyle(zoomLevel.DefaultTextStyle);
        }

        private bool IsDefaultAreaStyle(AreaStyle style)
        {
            if (!style.IsActive)
            {
                return true;
            }

            bool isDefault = false;

            if (style.OutlinePen.Color.IsTransparent && style.FillSolidBrush.Color.IsTransparent && style.OutlinePen.Brush is GeoSolidBrush)
            {
                isDefault = true;
            }

            if (isDefault && style.Advanced.FillCustomBrush != null)
            {
                isDefault = false;
            }

            if (isDefault && style.CustomAreaStyles.Count != 0)
            {
                isDefault = false;
            }

            return isDefault;
        }

        private bool IsDefaultLineStyle(LineStyle style)
        {
            if (!style.IsActive)
            {
                return true;
            }

            bool isDefault = false;

            if (style.OuterPen.Color.IsTransparent && style.InnerPen.Color.IsTransparent && style.CenterPen.Color.IsTransparent)
            {
                if (style.OuterPen.Brush is GeoSolidBrush && style.InnerPen.Brush is GeoSolidBrush && style.CenterPen.Brush is GeoSolidBrush)
                {
                    isDefault = true;
                }
            }

            if (style.CustomLineStyles.Count > 0)
            {
                isDefault = false;
            }

            return isDefault;
        }

        private bool IsDefaultPointStyle(PointStyle style)
        {
            if (!style.IsActive)
            {
                return true;
            }

            switch (style.PointType)
            {
                case PointType.Symbol:
                    bool isDefault = false;
                    if (style.SymbolPen.Color.IsTransparent && style.SymbolSolidBrush.Color.IsTransparent && style.SymbolPen.Brush is GeoSolidBrush)
                    {
                        isDefault = true;
                    }

                    if (isDefault && style.Advanced.CustomBrush != null)
                    {
                        isDefault = false;
                    }

                    return isDefault && style.CustomPointStyles.Count == 0;

                case PointType.Bitmap:
                    return style.Image == null && style.CustomPointStyles.Count == 0;

                case PointType.Character:
                    return style.CharacterSolidBrush.Color.IsTransparent && style.CustomPointStyles.Count == 0;
                default:
                    return true;
            }
        }

        private bool IsDefaultTextStyle(TextStyle style)
        {
            if (!style.IsActive)
            {
                return true;
            }

            bool isDefault = false;

            if (style.TextSolidBrush.Color.IsTransparent)
            {
                isDefault = true;
            }

            if (isDefault && style.Advanced.TextCustomBrush != null)
            {
                isDefault = false;
            }

            return isDefault;
        }

        private Feature GetInterSection(Feature feature, Collection<RectangleShape> boundingBoxes)
        {
            Feature result = null;

            foreach (var boundingBox in boundingBoxes)
            {
                var intersection = feature.GetIntersection(boundingBox.GetFeature());
                if (result == null)
                {
                    result = intersection;
                }
                else if (intersection != null)
                {
                    result = result.Union(intersection);
                }
            }

            return result;
        }

        private bool IsDisjointed(Feature feature, Collection<RectangleShape> boundingBoxes)
        {
            bool result = true;
            foreach (var boundingBox in boundingBoxes)
            {
                result = result && feature.IsDisjointed(boundingBox.GetFeature());
            }
            return result;
        }

        private Feature GetDifference(Feature feature, Collection<RectangleShape> boundingBoxes)
        {
            Feature result = null;
            foreach (var boundingBox in boundingBoxes)
            {
                if (result == null)
                {
                    result = SqlTypesGeometryHelper.GetDifference(feature, boundingBox.GetFeature());
                }
                else
                {
                    result = SqlTypesGeometryHelper.GetDifference(result, boundingBox.GetFeature());
                }
            }

            return result;
        }
    }
}
